from __future__ import annotations


__copyright__ = "Copyright (C) 2012-17 Andreas Kloeckner, Nick Curtis"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Mapping,
    Sequence,
    cast,
)

from constantdict import constantdict

from pymbolic import Variable, var
from pytools.codegen import CodeGenerator, Indentation
from pytools.py_codegen import PythonFunctionGenerator

from loopy.diagnostic import LoopyError


logger = logging.getLogger(__name__)

from pytools.persistent_dict import WriteOncePersistentDict

from loopy.kernel import KernelState, LoopKernel
from loopy.kernel.data import ArrayArg, _ArraySeparationInfo, auto
from loopy.tools import LoopyKeyBuilder, caches
from loopy.types import LoopyType, NumpyType
from loopy.typing import Expression, integer_expr_or_err
from loopy.version import DATA_MODEL_VERSION


if TYPE_CHECKING:
    from loopy.schedule.tools import KernelArgInfo
    from loopy.translation_unit import TranslationUnit


# {{{ object array argument packing

class SeparateArrayPackingController:
    """For argument arrays with axes tagged to be implemented as separate
    arrays, this class provides preprocessing of the incoming arguments so that
    all sub-arrays may be passed in one object array (under the original,
    un-split argument name) and are unpacked into separate arrays before being
    passed to the kernel.

    It also repacks outgoing arrays of this type back into an object array.
    """

    def __init__(self, packing_info: dict[str, _ArraySeparationInfo]) -> None:
        # These must work to index tuples if 1D.
        def untuple_length_1_indices(
                ind: tuple[int, ...]) -> int | tuple[int, ...]:
            if len(ind) == 1:
                return ind[0]
            else:
                return ind

        self.packing_info = {
                name: {
                    untuple_length_1_indices(ind): sep_name
                    for ind, sep_name in sep_info.subarray_names.items()
                    }
                for name, sep_info in packing_info.items()
                }

    def __call__(self, kernel_kwargs: dict[str, Any]) -> dict[str, Any]:
        kernel_kwargs = kernel_kwargs.copy()

        for name, ind_to_subary_name in self.packing_info.items():
            if name in kernel_kwargs:
                arg = kernel_kwargs[name]
                for index, unpacked_name in ind_to_subary_name.items():
                    assert unpacked_name not in kernel_kwargs
                    kernel_kwargs[unpacked_name] = arg[index]
                del kernel_kwargs[name]

        return kernel_kwargs

# }}}


# {{{ ExecutionWrapperGeneratorBase

def _str_to_expr(name_or_expr: str | Expression) -> Expression:
    if isinstance(name_or_expr, str):
        return var(name_or_expr)
    else:
        return name_or_expr


@dataclass(frozen=True)
class _ArgFindingEquation:
    lhs: Expression
    rhs: Expression

    # Arg finding code is sorted by priority, all equations (across all unknowns)
    # of lowest priority first.
    order: int

    based_on_names: frozenset[str]


class ExecutionWrapperGeneratorBase(ABC):
    """
    A set of common methods for generating a wrapper
    for execution

    """

    def __init__(self, system_args):
        self.system_args = system_args[:]

        from pytools import UniqueNameGenerator
        self.dtype_name_generator = UniqueNameGenerator(forced_prefix="_lpy_dtype_")
        self.dtype_str_to_name = {}

    @abstractmethod
    def python_dtype_str_inner(self, dtype):
        pass

    def python_dtype_str(self, gen: CodeGenerator, numpy_dtype):
        dtype_str = self.python_dtype_str_inner(numpy_dtype)
        try:
            return self.dtype_str_to_name[dtype_str]
        except KeyError:
            pass

        dtype_name = self.dtype_name_generator()
        gen.add_to_preamble(f"{dtype_name} = _lpy_np.dtype({dtype_str})")
        self.dtype_str_to_name[dtype_str] = dtype_name
        return dtype_name

    # {{{ invoker generation

    # /!\ This code runs in a namespace controlled by the user.
    # Prefix all auxiliary variables with "_lpy".

    # {{{ integer arg finding from array data

    def generate_integer_arg_finding_from_array_data(
            self, gen: CodeGenerator, kernel: LoopKernel, kai: KernelArgInfo
            ) -> None:
        from loopy.diagnostic import ParameterFinderWarning
        from loopy.kernel.array import get_strides
        from loopy.kernel.data import ArrayArg
        from loopy.symbolic import DependencyMapper, StringifyMapper
        dep_map: DependencyMapper[[]] = DependencyMapper()

        # {{{ find equations

        equations: list[_ArgFindingEquation] = []

        for arg_name in kai.passed_arg_names:
            arg = kernel.arg_dict[arg_name]
            assert arg.dtype is not None
            if isinstance(arg, ArrayArg):
                assert arg.shape is not auto
                if isinstance(arg.shape, tuple):
                    for axis_nr, shape_i in enumerate(arg.shape):
                        if shape_i is not None:
                            equations.append(
                                _ArgFindingEquation(
                                    lhs=var(arg.name).attr("shape")[axis_nr],
                                    rhs=shape_i,
                                    order=0,
                                    based_on_names=frozenset({arg.name})))

                strides = get_strides(arg)
                for axis_nr, stride_i in enumerate(strides):
                    if stride_i is not None:
                        equations.append(
                                _ArgFindingEquation(
                                    lhs=var("_lpy_even_div")(
                                        var(arg.name).attr("strides")[axis_nr],
                                        arg.dtype.itemsize),
                                    rhs=_str_to_expr(stride_i),
                                    order=0,
                                    based_on_names=frozenset({arg.name}),
                                    ))

                        if not arg.is_input and isinstance(arg.shape, tuple):
                            # If no value was found by other means, provide
                            # C-contiguous default strides for output-only
                            # arguments.
                            equations.append(
                                    _ArgFindingEquation(
                                        lhs=(integer_expr_or_err(strides[axis_nr + 1])
                                             * integer_expr_or_err(
                                                            arg.shape[axis_nr + 1]))
                                        if axis_nr + 1 < len(strides)
                                        else 1,
                                        rhs=_str_to_expr(stride_i),
                                        # Find strides from last dim to first,
                                        # starting at order=1 so that shape
                                        # parameters (found above) are
                                        # available.
                                        order=len(strides) - axis_nr,
                                        based_on_names=frozenset(),
                                        ))

                if arg.offset is not None:
                    equations.append(
                            _ArgFindingEquation(
                                lhs=var("_lpy_even_div_none")(
                                    var("getattr")(
                                        var(arg.name), var('"offset"'), var("None")),
                                    arg.dtype.itemsize),
                                rhs=_str_to_expr(arg.offset),
                                order=0,
                                based_on_names=frozenset([arg.name]),
                                ))

                    # If no value was found by other means, default to zero.
                    equations.append(
                            _ArgFindingEquation(
                                lhs=0,
                                rhs=_str_to_expr(arg.offset),
                                order=1,
                                based_on_names=frozenset(),
                                ))

        # }}}

        # {{{ regroup equations by unknown

        order_to_unknown_to_equations: \
                dict[int, dict[str, list[_ArgFindingEquation]]] = {}

        for eqn in equations:
            deps = dep_map(eqn.rhs)

            if len(deps) == 1:
                unknown_var, = deps
                order_to_unknown_to_equations \
                        .setdefault(eqn.order, {}) \
                        .setdefault(cast("Variable", unknown_var).name, []) \
                        .append(eqn)
            else:
                # Zero deps: nothing to determine, forget about it.
                # 2+ deps: not implemented
                pass

        del equations

        # }}}

        # {{{ generate arg finding code

        from pymbolic.algorithm import solve_affine_equations_for
        from pytools.codegen import CodeGenerator

        gen("# {{{ find integer arguments from array data")
        gen("")

        for order_value in sorted(order_to_unknown_to_equations):
            for unknown_name in sorted(order_to_unknown_to_equations[order_value]):
                unk_equations = sorted(
                        order_to_unknown_to_equations[order_value][unknown_name],
                        key=lambda eqn: eqn.order)
                subgen = CodeGenerator()

                seen_based_on_names: set[frozenset[str]] = set()

                if_or_elif = "if"

                for eqn in unk_equations:
                    if eqn.rhs == Variable(unknown_name):
                        # Some of the expressions above are non-affine. Let's not
                        # get carried away by trying to solve a much more complex
                        # problem than needed.
                        value_expr = eqn.lhs
                    else:
                        try:
                            # overkill :)
                            value_expr = solve_affine_equations_for(
                                    [unknown_name],
                                    [(eqn.lhs, eqn.rhs)]
                                    )[Variable(unknown_name)]
                        except Exception as e:
                            # went wrong? oh well
                            from warnings import warn
                            warn("Unable to generate code to automatically "
                                    f"find '{unknown_name}' "
                                    f"from '{', '.join(eqn.based_on_names)}':\n"
                                    f"{e}", ParameterFinderWarning, stacklevel=1)
                            continue

                    # Do not use more than one bit of data from each of the
                    # 'based_on_names' to find each value, i.e. if a value can be
                    # found via shape and strides, only one of them suffices.
                    # This also helps because strides can be unreliable in the
                    # face of zero-length axes.
                    if eqn.based_on_names in seen_based_on_names:
                        continue
                    seen_based_on_names.add(eqn.based_on_names)

                    if eqn.based_on_names:
                        condition = " and ".join(
                                f"{ary_name} is not None"
                                for ary_name in eqn.based_on_names)
                    else:
                        condition = "True"

                    subgen(f"{if_or_elif} {condition}:")
                    with Indentation(subgen):
                        subgen(
                                f"{unknown_name} = {StringifyMapper()(value_expr)}")
                    if_or_elif = "elif"

                    subgen("")

                if subgen.code:
                    gen(f"if {unknown_name} is None:")
                    with Indentation(gen):
                        gen.extend(subgen)

        gen("# }}}")
        gen("")

        # }}}

    # }}}

    # {{{ check that value args are present

    def generate_value_arg_check(
            self, gen: CodeGenerator, kernel: LoopKernel, kai: KernelArgInfo
            ) -> None:
        if kernel.options.skip_arg_checks:
            return

        from loopy.kernel.data import ValueArg

        gen("# {{{ check that value args are present")
        gen("")

        for arg_name in kai.passed_arg_names:
            arg = kernel.arg_dict[arg_name]
            if not isinstance(arg, ValueArg):
                continue

            gen("if %s is None:" % arg.name)
            with Indentation(gen):
                gen("raise TypeError(\"value argument '%s' "
                        "was not given and could not be automatically "
                        'determined")' % arg.name)

        gen("# }}}")
        gen("")

    # }}}

    # {{{ handle non numpy arguments

    def handle_non_numpy_arg(self, gen: CodeGenerator, arg):
        raise NotImplementedError()

    # }}}

    # {{{ handle allocation of unspecified arguments

    def handle_alloc(
            self, gen: CodeGenerator, arg: ArrayArg,
            strify: Callable[[Expression | tuple[Expression]], str],
            skip_arg_checks: bool) -> None:
        """
        Handle allocation of non-specified arguments for C-execution
        """
        raise NotImplementedError()

    # }}}

    def get_arg_pass(self, arg):
        raise NotImplementedError()

    def get_strides_check_expr(self, shape, strides, expected_strides):
        assert len(shape) == len(strides) == len(expected_strides)

        # Returns an expression suitable for use for checking the strides of an
        # argument. Arguments should be sequences of strings.

        # Shape axes of length 1 are ignored because strides along these
        # axes are never used: The only valid index is 1.
        match_expr = " and ".join(
                f"({shape_i} == 1 or {strides_i} == {expected_strides_i})"
                for shape_i, strides_i, expected_strides_i
                in zip(shape, strides, expected_strides)) or "True"

        if shape:
            # If any shape component is zero, the array is empty and the strides
            # don't matter.
            match_expr = (f"({match_expr})"
            + "".join(f" or not {shape_i}" for shape_i in shape))

        return match_expr

    # {{{ arg setup

    def generate_arg_setup(
            self, gen: CodeGenerator, kernel: LoopKernel, kai: KernelArgInfo,
            ) -> Sequence[str]:
        options = kernel.options
        import loopy as lp
        from loopy.kernel.array import ArrayBase
        from loopy.kernel.data import ImageArg
        from loopy.symbolic import StringifyMapper
        from loopy.types import NumpyType

        gen("# {{{ set up array arguments")
        gen("")

        if not options.no_numpy:
            gen("_lpy_encountered_numpy = False")
            gen("_lpy_encountered_dev = False")
            gen("")

        args = []

        strify = StringifyMapper()

        for arg_name in kai.passed_arg_names:
            arg = kernel.arg_dict[arg_name]
            is_written = arg.name in kernel.get_written_variables()

            if not isinstance(arg, ArrayBase):
                args.append(arg.name)
                continue

            gen("# {{{ process %s" % arg.name)
            gen("")

            if not options.no_numpy:
                self.handle_non_numpy_arg(gen, arg)

            if not options.skip_arg_checks and arg.is_input:
                gen("if %s is None:" % arg.name)
                with Indentation(gen):
                    gen("raise RuntimeError(\"input argument '%s' must "
                            'be supplied")' % arg.name)
                    gen("")

            if (is_written and isinstance(arg, ImageArg)
                    and not options.skip_arg_checks):
                gen("if %s is None:" % arg.name)
                with Indentation(gen):
                    gen("raise RuntimeError(\"written image '%s' must "
                            'be supplied")' % arg.name)
                    gen("")

            if is_written and arg.shape is None and not options.skip_arg_checks:
                gen("if %s is None:" % arg.name)
                with Indentation(gen):
                    gen("raise RuntimeError(\"written argument '%s' has "
                            'unknown shape and must be supplied")' % arg.name)
                    gen("")

            possibly_made_by_loopy = False

            # {{{ allocate written arrays, if needed

            if arg.is_output \
                    and isinstance(arg, (lp.ArrayArg, lp.ConstantArg)) \
                    and arg.shape is not None \
                    and all(si is not None for si in arg.shape):

                if not isinstance(arg.dtype, NumpyType):
                    raise LoopyError("do not know how to pass arg of type '%s'"
                            % arg.dtype)

                possibly_made_by_loopy = True
                gen("_lpy_made_by_loopy = False")
                gen("")

                gen("if %s is None:" % arg.name)
                with Indentation(gen):
                    self.handle_alloc(
                        gen, arg, strify, options.skip_arg_checks)
                    gen("_lpy_made_by_loopy = True")
                    gen("")

            # }}}

            # {{{ argument checking

            if isinstance(arg, (lp.ArrayArg, lp.ConstantArg)) \
                    and not options.skip_arg_checks:
                if possibly_made_by_loopy:
                    gen("if not _lpy_made_by_loopy:")
                else:
                    gen("if True:")

                with Indentation(gen):
                    gen("if %s.dtype != %s:"
                            % (arg.name, self.python_dtype_str(
                                gen, arg.dtype.numpy_dtype)))
                    with Indentation(gen):
                        gen("raise TypeError(\"dtype mismatch on argument '%s' "
                                '(got: %%s, expected: %s)" %% %s.dtype)'
                                % (arg.name, arg.dtype, arg.name))

                    # {{{ generate shape checking code

                    def strify_allowing_none(shape_axis):
                        if shape_axis is None:
                            return "None"
                        else:
                            return strify(shape_axis)

                    def strify_tuple(t: tuple[Expression, ...] | None) -> str:
                        if t is None:
                            return "None"
                        if len(t) == 0:
                            return "()"
                        else:
                            return "(%s,)" % ", ".join(
                                    strify_allowing_none(sa)
                                    for sa in t)

                    shape_mismatch_msg = (
                            "raise ValueError(\"shape mismatch on argument '%s' "
                            '(got: %%s, expected: %%s)" '
                            "%% (%s.shape, %s))"
                            % (arg.name, arg.name, strify_tuple(arg.shape)))

                    if arg.shape is None:
                        pass

                    elif any(shape_axis is None for shape_axis in arg.shape):
                        assert isinstance(arg.shape, tuple)
                        gen("if len(%s.shape) != %s:"
                                % (arg.name, len(arg.shape)))
                        with Indentation(gen):
                            gen(shape_mismatch_msg)

                        for i, shape_axis in enumerate(arg.shape):
                            if shape_axis is None:
                                continue

                            gen("if %s.shape[%d] != %s:"
                                    % (arg.name, i, strify(shape_axis)))
                            with Indentation(gen):
                                gen(shape_mismatch_msg)

                    else:  # not None, no Nones in tuple
                        gen("if %s.shape != %s:"
                                % (arg.name, strify(arg.shape)))
                        with Indentation(gen):
                            gen(shape_mismatch_msg)

                    # }}}

                    from loopy.kernel.array import get_strides
                    strides = get_strides(arg)

                    if strides and arg.dim_tags and arg.shape is not None:
                        assert isinstance(arg.shape, tuple)
                        itemsize = arg.dtype.numpy_dtype.itemsize
                        sym_strides = tuple(itemsize*s_i for s_i in strides)

                        ndim = len(arg.shape)
                        shape = ["_lpy_shape_%d" % i for i in range(ndim)]
                        strides = ["_lpy_stride_%d" % i for i in range(ndim)]

                        gen("({},) = {}.shape".format(", ".join(shape), arg.name))
                        gen("({},) = {}.strides".format(
                            ", ".join(strides), arg.name))

                        gen("if not (%s):"
                                % self.get_strides_check_expr(
                                    shape, strides,
                                    [strify(s) for s in sym_strides]))
                        with Indentation(gen):
                            gen(f"_lpy_got = {arg.name}.strides")
                            gen(f"_lpy_expected = {strify_tuple(sym_strides)}")

                            gen('raise ValueError("strides mismatch on '
                                    "argument '%s' "
                                    '(got: %%s, expected: %%s)" '
                                    "%% (_lpy_got, _lpy_expected))"
                                    % arg.name)

                    if not arg.offset:
                        gen("if hasattr({}, 'offset') and {}.offset:".format(
                                arg.name, arg.name))
                        with Indentation(gen):
                            gen("raise ValueError(\"Argument '%s' does not "
                                    "allow arrays with offsets. Try passing "
                                    "default_offset=loopy.auto to make_kernel()."
                                    '")' % arg.name)
                            gen("")

            # }}}

            if possibly_made_by_loopy and not options.skip_arg_checks:
                gen("del _lpy_made_by_loopy")
                gen("")

            if isinstance(arg, (lp.ArrayArg, lp.ConstantArg)):
                args.append(self.get_arg_pass(arg))
            else:
                args.append(arg.name)

            gen("")

            gen("# }}}")
            gen("")

        gen("# }}}")
        gen("")

        return args

    # }}}

    def target_specific_preamble(self, gen):
        """
        Add target specific imports to preamble
        """
        raise NotImplementedError()

    def initialize_system_args(self, gen):
        """
        Override to initialize any default system args
        """
        raise NotImplementedError()

    # {{{ generate invocation

    def generate_invocation(self, gen: CodeGenerator, kernel: LoopKernel,
            kai: KernelArgInfo, host_program_name: str, args: Sequence[str]) -> None:
        raise NotImplementedError()

    # }}}

    # {{{ output

    def generate_output_handler(self, gen: CodeGenerator,
            kernel: LoopKernel, kai: KernelArgInfo) -> None:
        raise NotImplementedError()

    # }}}

    def generate_host_code(self, gen, codegen_result):
        raise NotImplementedError

    def __call__(self, program, entrypoint, codegen_result):
        """
        Generates the wrapping python invoker for this execution target

        :arg kernel: the loopy :class:`LoopKernel`(s) to be executed
        :codegen_result: the loopy :class:`CodeGenerationResult` created
        by code generation

        :returns: A python callable that handles execution of this
            kernel
        """

        kernel = program[entrypoint]
        options = kernel.options

        from loopy.schedule.tools import get_kernel_arg_info
        kai = get_kernel_arg_info(kernel)

        gen = PythonFunctionGenerator(
                "invoke_%s_loopy_kernel" % entrypoint,
                self.system_args + [
                    "%s=None" % arg_name
                    for arg_name in kai.passed_arg_names
                    ])

        self.target_specific_preamble(gen)
        gen.add_to_preamble("")
        self.generate_host_code(gen, codegen_result)
        gen.add_to_preamble("")

        self.initialize_system_args(gen)

        self.generate_integer_arg_finding_from_array_data(
                gen, program[entrypoint], kai)
        self.generate_value_arg_check(gen, program[entrypoint], kai)
        args = self.generate_arg_setup(gen, program[entrypoint], kai)

        # FIXME: should we make this as a dict as well.
        host_program_name = codegen_result.host_programs[entrypoint].name

        self.generate_invocation(gen, program[entrypoint], kai,
                host_program_name, args)

        self.generate_output_handler(gen, program[entrypoint], kai)

        if options.write_wrapper:
            output = gen.get()
            if options.allow_terminal_colors:
                output = get_highlighted_python_code(output)

            if options.write_wrapper is True:
                print(output)
            else:
                with open(options.write_wrapper, "w") as outf:
                    outf.write(output)

        return gen.get_picklable_function()

# }}}

# }}}


typed_and_scheduled_cache: WriteOncePersistentDict[
    tuple[str, TranslationUnit, Mapping[str, LoopyType] | None],
    TranslationUnit
] = WriteOncePersistentDict(
        "loopy-typed-and-scheduled-cache-v1-"+DATA_MODEL_VERSION,
        key_builder=LoopyKeyBuilder(),
        safe_sync=False)


caches.append(typed_and_scheduled_cache)


invoker_cache: WriteOncePersistentDict[
    tuple[str, TranslationUnit, str],
    str
] = WriteOncePersistentDict(
        "loopy-invoker-cache-v10-"+DATA_MODEL_VERSION,
        key_builder=LoopyKeyBuilder(),
        safe_sync=False)


caches.append(invoker_cache)


# {{{ kernel executor

class ExecutorBase:
    """An object allowing the execution of an entrypoint of a
    :class:`~loopy.TranslationUnit`. Create these objects using
    :meth:`loopy.TranslationUnit.executor`.

    .. automethod:: __call__
    """
    packing_controller: SeparateArrayPackingController | None

    def __init__(self, t_unit: TranslationUnit, entrypoint: str):
        self.t_unit = t_unit
        self.entrypoint = entrypoint

        kernel = self.t_unit[entrypoint]
        self.output_names = {arg.name for arg in kernel.args if arg.is_output}

        from loopy import ArrayArg
        self.input_array_names = {
            arg.name for arg in kernel.args
            if arg.is_input and isinstance(arg, ArrayArg)}

        self.has_runtime_typed_args = any(arg.dtype is None for arg in kernel.args)

        # We're doing this ahead of time to learn about array separation.
        # This will be done again as part of preprocessing below, and we're
        # betting that it happens consistently both times. (No reason it wouldn't,
        # but it is done redundantly.) We can't *use* the result of this
        # because we need to do the 'official' array separation after type
        # inference has completed.
        from loopy.preprocess import make_arrays_for_sep_arrays
        self.separated_entry_knl = make_arrays_for_sep_arrays(
                self.t_unit[self.entrypoint])

        self.sep_info = self.separated_entry_knl._separation_info()
        if self.sep_info:
            self.packing_controller = SeparateArrayPackingController(self.sep_info)
        else:
            self.packing_controller = None

    def check_for_required_array_arguments(self, input_args):
        # Formerly, the first exception raised when a required argument is not
        # passed was often at type inference. This exists to raise a more meaningful
        # message in such scenarios. Since type inference precedes compilation, this
        # check cannot be deferred to the generated invoker code.
        # See discussion at
        # https://github.com/inducer/loopy/pull/160#issuecomment-867761204
        # and links therin for context.
        if not self.input_array_names <= set(input_args):
            missing_args = self.input_array_names - set(input_args)
            kernel = self.t_unit[self.entrypoint]
            raise LoopyError(
                f"Kernel {kernel.name}() missing required array input arguments: "
                f"{', '.join(missing_args)}. "
                "If this is a surprise, maybe you need to add is_input=False to "
                "your argument.")

    def get_typed_and_scheduled_translation_unit_uncached(
            self, arg_to_dtype: constantdict[str, LoopyType] | None
            ) -> TranslationUnit:
        t_unit = self.t_unit

        if arg_to_dtype:
            entry_knl = t_unit[self.entrypoint]

            # FIXME: This is not so nice. This transfers types from the
            # subarrays of sep-tagged arrays to the 'main' array, because
            # type inference fails otherwise.
            mm = arg_to_dtype.mutate()
            for name, sep_info in self.sep_info.items():
                if entry_knl.arg_dict[name].dtype is None:
                    for sep_name in sep_info.subarray_names.values():
                        if sep_name in arg_to_dtype:
                            mm[name] = arg_to_dtype[sep_name]
                            del mm[sep_name]

            arg_to_dtype = mm.finish()

            from loopy.kernel.tools import add_dtypes
            t_unit = t_unit.with_kernel(add_dtypes(entry_knl, arg_to_dtype))

            from loopy.type_inference import infer_unknown_types
            t_unit = infer_unknown_types(t_unit, expect_completion=True)

        if t_unit.state < KernelState.PREPROCESSED:
            from loopy.preprocess import preprocess_program
            t_unit = preprocess_program(t_unit)

        if t_unit.state < KernelState.LINEARIZED:
            from loopy.schedule import linearize
            t_unit = linearize(t_unit)

        return t_unit

    def get_typed_and_scheduled_translation_unit(
            self, arg_to_dtype: constantdict[str, LoopyType] | None
            ) -> TranslationUnit:
        from loopy import CACHING_ENABLED

        cache_key = (type(self).__name__, self.t_unit, arg_to_dtype)

        if CACHING_ENABLED:
            try:
                return typed_and_scheduled_cache[cache_key]
            except KeyError:
                pass

        logger.debug("%s: typed-and-scheduled cache miss" %
                self.t_unit.entrypoints)

        t_unit = self.get_typed_and_scheduled_translation_unit_uncached(arg_to_dtype)

        if CACHING_ENABLED:
            typed_and_scheduled_cache.store_if_not_present(cache_key, t_unit)

        return t_unit

    def arg_to_dtype(self, kwargs) -> constantdict[str, LoopyType] | None:
        if not self.has_runtime_typed_args:
            return None

        arg_dict = self.separated_entry_knl.arg_dict
        arg_to_dtype = {}
        for arg_name, val in kwargs.items():
            arg = arg_dict[arg_name]

            if arg.dtype is None and val is not None:
                try:
                    dtype = val.dtype
                except AttributeError:
                    pass
                else:
                    arg_to_dtype[arg_name] = NumpyType(dtype)

        return constantdict(arg_to_dtype)

    # {{{ debugging aids

    def get_highlighted_code(self, entrypoint, arg_to_dtype=None, code=None):
        if code is None:
            code = self.get_code(entrypoint, arg_to_dtype)
        return get_highlighted_code(code)

    def get_code(
            self, entrypoint: str,
            arg_to_dtype: constantdict[str, LoopyType] | None = None) -> str:
        kernel = self.get_typed_and_scheduled_translation_unit(arg_to_dtype)

        from loopy.codegen import generate_code_v2
        code = generate_code_v2(kernel)
        return code.device_code()

    def get_invoker_uncached(self, program, entrypoint, *args):
        raise NotImplementedError()

    def get_invoker(self, t_unit, entrypoint, *args):
        from loopy import CACHING_ENABLED

        cache_key = (self.__class__.__name__, (t_unit, entrypoint))

        if CACHING_ENABLED:
            try:
                return invoker_cache[cache_key]
            except KeyError:
                pass

        logger.debug("%s: invoker cache miss" % entrypoint)

        invoker = self.get_invoker_uncached(t_unit, entrypoint, *args)

        if CACHING_ENABLED:
            invoker_cache.store_if_not_present(cache_key, invoker)

        return invoker

    # }}}

    # {{{ call and info generator

    def __call__(self, queue, **kwargs):
        raise NotImplementedError()

    # }}}

# }}}

# {{{ code highlighters


def get_highlighted_code(text, python=False):
    try:
        from pygments import highlight
    except ImportError:
        return text
    else:
        from pygments.formatters import TerminalFormatter
        from pygments.lexers import CLexer, PythonLexer

        return highlight(text, CLexer() if not python else PythonLexer(),
                         TerminalFormatter())


def get_highlighted_python_code(text):
    return get_highlighted_code(text, True)


# }}}

# vim: foldmethod=marker



import numpy as np
from constantdict import constantdict

import loopy as lp


# {{{ test_barrier_in_overridden_get_grid_size_expanded_kernel

class GridOverride:
    def __init__(self, clean, vecsize):
        self.clean = clean
        self.vecsize = vecsize

    def __call__(self, insn_ids, callables_table, ignore_auto=True):
        gsize, _ = self.clean.get_grid_sizes_for_insn_ids(insn_ids,
                callables_table, ignore_auto)
        return gsize, (self.vecsize,)

# }}}


# {{{ test_register_function_lookup

class Log2Callable(lp.ScalarCallable):

    def with_types(self, arg_id_to_dtype, callables_table):

        if 0 not in arg_id_to_dtype or arg_id_to_dtype[0] is None:
            # the types provided aren't mature enough to specialize the
            # callable
            return (
                    self.copy(arg_id_to_dtype=constantdict(arg_id_to_dtype)),
                    callables_table)

        dtype = arg_id_to_dtype[0].numpy_dtype

        if dtype.kind in ("u", "i"):
            # ints and unsigned casted to float32
            dtype = np.float32

        if dtype.type == np.float32:
            name_in_target = "log2f"
        elif dtype.type == np.float64:
            name_in_target = "log2"
            pass
        else:
            raise TypeError(f"log2: unexpected type {dtype}")

        from loopy.types import NumpyType
        return (
                self.copy(name_in_target=name_in_target,
                    arg_id_to_dtype=constantdict({
                        0: NumpyType(dtype),
                        -1: NumpyType(dtype)
                        })),
                callables_table)


# }}}

# vim: foldmethod=marker



"""Python host AST builder for integration with PyOpenCL."""
from __future__ import annotations


__copyright__ = "Copyright (C) 2016 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from typing import TYPE_CHECKING, Sequence

import numpy as np

from genpy import Collection, Generable, Suite
from pymbolic.mapper import Mapper
from pymbolic.mapper.stringifier import StringifyMapper

from loopy.diagnostic import LoopyError
from loopy.kernel.data import ValueArg
from loopy.target import ASTBuilderBase
from loopy.type_inference import TypeReader


if TYPE_CHECKING:
    from loopy.codegen import CodeGenerationState
    from loopy.codegen.result import CodeGenerationResult


# {{{ expression to code

class ExpressionToPythonMapper(StringifyMapper):
    def __init__(self, codegen_state, type_inf_mapper=None):
        self.kernel = codegen_state.kernel
        self.codegen_state = codegen_state

        if type_inf_mapper is None:
            type_inf_mapper = TypeReader(self.kernel,
                    self.codegen_state.callables_table)
        self.type_inf_mapper = type_inf_mapper

    def handle_unsupported_expression(self, victim, enclosing_prec):
        return Mapper.handle_unsupported_expression(self, victim, enclosing_prec)

    def rec(self, expr, prec, type_context=None, needed_dtype=None):
        return super().rec(expr, prec)

    # FIXME: Fix once mappers are precisely typed
    __call__ = rec  # type: ignore[assignment]

    def map_constant(self, expr, enclosing_prec):
        if isinstance(expr, np.generic):
            return repr(expr).replace("np.", "_lpy_np.")
        else:
            return repr(expr)

    def map_variable(self, expr, enclosing_prec):
        if expr.name in self.codegen_state.var_subst_map:
            # Unimplemented: annotate_inames
            return str(self.rec(
                self.codegen_state.var_subst_map[expr.name],
                enclosing_prec))

        if expr.name in self.kernel.all_inames():
            return super().map_variable(
                    expr, enclosing_prec)

        var_descr = self.kernel.get_var_descriptor(expr.name)
        if isinstance(var_descr, ValueArg):
            return super().map_variable(
                    expr, enclosing_prec)

        return super().map_variable(
                expr, enclosing_prec)

    def map_subscript(self, expr, enclosing_prec):
        return super().map_subscript(
                expr, enclosing_prec)

    def map_call(self, expr, enclosing_prec):
        from pymbolic.mapper.stringifier import PREC_NONE

        identifier_name = self.codegen_state.callables_table[
                expr.function.name].name

        if identifier_name in ["indexof", "indexof_vec"]:
            raise LoopyError(
                    "indexof, indexof_vec not yet supported in Python")

        clbl = self.codegen_state.callables_table[
                expr.function.name]

        str_parameters = None
        number_of_assignees = len([key for key in
            clbl.arg_id_to_dtype.keys() if key < 0])

        if number_of_assignees != 1:
            raise LoopyError("functions with more or fewer than one return value "
                    "may not be used in an expression")

        str_parameters = [self.rec(par, PREC_NONE) for par in expr.parameters]

        return "{}({})".format(clbl.name_in_target,
                               ", ".join(str_parameters))

    def map_group_hw_index(self, expr, enclosing_prec):
        raise LoopyError("plain Python does not have group hw axes")

    def map_local_hw_index(self, expr, enclosing_prec):
        raise LoopyError("plain Python does not have local hw axes")

    def map_if(self, expr, enclosing_prec):
        # Synthesize PREC_IFTHENELSE, make sure it is in the right place in the
        # operator precedence hierarchy (right above "or").
        from pymbolic.mapper.stringifier import PREC_LOGICAL_OR
        PREC_IFTHENELSE = PREC_LOGICAL_OR - 1  # noqa

        return self.parenthesize_if_needed(
            "{then} if {cond} else {else_}".format(
                # "1 if 0 if 1 else 2 else 3" is not valid Python.
                # So force parens by using an artificially higher precedence.
                then=self.rec(expr.then, PREC_LOGICAL_OR),
                cond=self.rec(expr.condition, PREC_LOGICAL_OR),
                else_=self.rec(expr.else_, PREC_LOGICAL_OR)),
            enclosing_prec, PREC_IFTHENELSE)

# }}}


# {{{ ast builder

def _base_python_preamble_generator(preamble_info):
    yield ("00_future", "from __future__ import division, print_function\n")
    yield ("05_numpy_import", """
            import numpy as _lpy_np
            """)


class PythonASTBuilderBase(ASTBuilderBase[Generable]):
    """A Python host AST builder for integration with PyOpenCL.
    """

    @property
    def known_callables(self):
        from loopy.target.c import get_c_callables
        callables = super().known_callables
        callables.update(get_c_callables())
        return callables

    def preamble_generators(self):
        return (
                [*super().preamble_generators(), _base_python_preamble_generator])

    # {{{ code generation guts

    @property
    def ast_module(self):
        import genpy
        return genpy

    def get_function_declaration(
            self, codegen_state: CodeGenerationState,
            codegen_result: CodeGenerationResult, schedule_index: int
            ) -> tuple[Sequence[tuple[str, str]], Generable | None]:
        return [], None

    def get_function_definition(self, codegen_state, codegen_result,
            schedule_index,
            function_decl, function_body):

        assert function_decl is None

        from genpy import Function
        return Function(
                codegen_result.current_program(codegen_state).name,
                [idi.name for idi in codegen_state.implemented_data_info],
                function_body)

    def get_temporary_decls(self, codegen_state, schedule_index):
        kernel = codegen_state.kernel
        ecm = codegen_state.expression_to_code_mapper

        result = []

        from genpy import Assign
        from pymbolic.mapper.stringifier import PREC_NONE

        for tv in sorted(
                kernel.temporary_variables.values(),
                key=lambda key_tv: key_tv.name):
            if tv.shape:
                result.append(
                        Assign(
                            tv.name,
                            "_lpy_np.empty(%s, dtype=%s)"
                            % (
                                ecm(tv.shape, PREC_NONE, "i"),
                                "_lpy_np."+(
                                    tv.dtype.numpy_dtype.name
                                    if tv.dtype.numpy_dtype.name != "bool"
                                    else "bool_")
                                )))

        return result

    def get_expression_to_code_mapper(self, codegen_state):
        return ExpressionToPythonMapper(codegen_state)

    @property
    def ast_block_class(self):
        return Suite

    @property
    def ast_block_scope_class(self):
        # Once a new version of genpy is released, switch to this:
        # from genpy import Collection
        # and delete the implementation above.
        return Collection

    def emit_sequential_loop(self, codegen_state, iname, iname_dtype,
            lbound, ubound, inner, hints):
        ecm = codegen_state.expression_to_code_mapper

        from genpy import For
        from pymbolic.mapper.stringifier import PREC_NONE, PREC_SUM

        if hints:
            raise ValueError("hints for python loops not supported")

        return For(
                (iname,),
                "range(%s, %s + 1)"
                % (
                    ecm(lbound, PREC_NONE, "i"),
                    ecm(ubound, PREC_SUM, "i"),
                    ),
                inner)

    def emit_initializer(self, codegen_state, dtype, name, val_str, is_const):
        from genpy import Assign
        return Assign(name, val_str)

    def emit_blank_line(self):
        from genpy import Line
        return Line()

    def emit_comment(self, s):
        from genpy import Comment
        return Comment(s)

    def emit_noop_with_comment(self, s):
        from cgen import Line
        return Line(f"pass #{s}")

    @property
    def can_implement_conditionals(self):
        return True

    def emit_if(self, condition_str, ast):
        from genpy import If
        return If(condition_str, ast)

    def emit_assignment(self, codegen_state, insn):
        ecm = codegen_state.expression_to_code_mapper

        if insn.atomicity:
            raise NotImplementedError("atomic ops in Python")

        from genpy import Assign
        from pymbolic.mapper.stringifier import PREC_NONE

        return Assign(
                ecm(insn.assignee, prec=PREC_NONE, type_context=None),
                ecm(insn.expression, prec=PREC_NONE, type_context=None))

    # }}}

# }}}

# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from typing import TYPE_CHECKING

import numpy as np
from constantdict import constantdict

from pymbolic import var
from pymbolic.primitives import expr_dataclass

from loopy.diagnostic import LoopyError
from loopy.kernel.function_interface import ScalarCallable
from loopy.symbolic import FunctionIdentifier, ResolvedFunction
from loopy.tools import update_persistent_hash
from loopy.types import NumpyType


if TYPE_CHECKING:
    from pytools.persistent_dict import Hash, KeyBuilder


__doc__ = """
.. currentmodule:: loopy.library.reduction

.. autoclass:: ReductionOperation

.. autoclass:: ScalarReductionOperation

.. autoclass:: SumReductionOperation

.. autoclass:: ProductReductionOperation

.. autoclass:: MaxReductionOperation

.. autoclass:: MinReductionOperation

.. autoclass:: ReductionOpFunction
"""


class ReductionOperation:
    """Subclasses of this type have to be hashable, picklable, and
    equality-comparable.
    """

    def result_dtypes(self, *arg_dtypes):
        """
        :arg arg_dtypes: may be None if not known
        :returns: None if not known, otherwise the returned type
        """

        raise NotImplementedError

    @property
    def arg_count(self):
        raise NotImplementedError

    def neutral_element(self, dtypes, callables_table, target):
        raise NotImplementedError

    def __hash__(self):
        # Force subclasses to override
        raise NotImplementedError

    def __eq__(self, other):
        # Force subclasses to override
        raise NotImplementedError

    def __call__(self, dtype, operand1, operand2):
        raise NotImplementedError

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return type(self).__name__

    @staticmethod
    def parse_result_type(target, op_type):
        try:
            return NumpyType(np.dtype(op_type))
        except TypeError:
            pass

        if op_type.startswith("vec_"):
            try:
                return NumpyType(target.get_or_register_dtype(op_type[4:]))
            except AttributeError:
                pass

        raise LoopyError("unable to parse reduction type: '%s'"
                % op_type)


class ScalarReductionOperation(ReductionOperation):
    @property
    def arg_count(self):
        return 1

    def result_dtypes(self, arg_dtype):
        if arg_dtype is None:
            return (None,)

        return (arg_dtype,)

    def __hash__(self):
        return hash((type(self),))

    def __eq__(self, other):
        return type(self) is type(other)

    def __str__(self):
        result = type(self).__name__.replace("ReductionOperation", "").lower()

        return result

    def update_persistent_hash(self, key_hash: Hash, key_builder: KeyBuilder) -> None:
        # They're all stateless.
        key_builder.rec(key_hash, type(self))


class SumReductionOperation(ScalarReductionOperation):
    def neutral_element(self, dtype, callables_table, target):
        # FIXME: Document that we always use an int here.
        from loopy import auto
        if dtype not in [None, auto] and dtype.numpy_dtype.kind == "f":
            return 0.0, callables_table

        return 0, callables_table

    def __call__(self, dtype, operand1, operand2, callables_table, target):
        return operand1 + operand2, callables_table


class ProductReductionOperation(ScalarReductionOperation):
    def neutral_element(self, dtype, callables_table, target):
        # FIXME: Document that we always use an int here.
        from loopy import auto
        if dtype not in [None, auto] and dtype.numpy_dtype.kind == "f":
            return 1.0, callables_table

        return 1, callables_table

    def __call__(self, dtype, operand1, operand2, callables_table, target):
        return operand1 * operand2, callables_table


class AnyReductionOperation(ScalarReductionOperation):
    def neutral_element(self, dtype, callables_table, target):
        return False, callables_table

    def __call__(self, dtype, operand1, operand2, callables_table, target):
        from pymbolic.primitives import LogicalOr
        return LogicalOr((operand1, operand2)), callables_table


class AllReductionOperation(ScalarReductionOperation):
    def neutral_element(self, dtype, callables_table, target):
        return True, callables_table

    def __call__(self, dtype, operand1, operand2, callables_table, target):
        from pymbolic.primitives import LogicalAnd
        return LogicalAnd((operand1, operand2)), callables_table


def get_le_neutral(dtype):
    """Return a number y that satisfies (x <= y) for all y."""

    if dtype.numpy_dtype.kind == "f":
        # OpenCL 1.2, section 6.12.2
        if dtype.numpy_dtype.itemsize == 4:
            # float
            return var("INFINITY")
        elif dtype.numpy_dtype.itemsize == 8:
            # double
            return var("HUGE_VAL")

    elif dtype.numpy_dtype.kind == "i":
        # OpenCL 1.1, section 6.11.3
        if dtype.numpy_dtype.itemsize == 4:
            # 32 bit integer
            return var("INT_MAX")
        elif dtype.numpy_dtype.itemsize == 8:
            # 64 bit integer
            return var("LONG_MAX")
    elif dtype.numpy_dtype.kind == "u":
        if dtype.numpy_dtype.itemsize == 4:
            # 32 bit integer
            return var("UINT_MAX")
        elif dtype.numpy_dtype.itemsize == 8:
            # 64 bit integer
            return var("ULONG_MAX")

    raise NotImplementedError(f"neutral element for <= and {dtype}")


def get_ge_neutral(dtype):
    """Return a number y that satisfies (x >= y) for all y."""

    if dtype.numpy_dtype.kind == "f":
        # OpenCL 1.2, section 6.12.2
        if dtype.numpy_dtype.itemsize == 4:
            # float
            return -var("INFINITY")
        elif dtype.numpy_dtype.itemsize == 8:
            # double
            return -var("HUGE_VAL")
    elif dtype.numpy_dtype.kind == "i":
        # OpenCL 1.1, section 6.11.3
        if dtype.numpy_dtype.itemsize == 4:
            # 32 bit integer
            return var("INT_MIN")
        elif dtype.numpy_dtype.itemsize == 8:
            # 64 bit integer
            return var("LONG_MIN")
    elif dtype.numpy_dtype.kind == "u":
        return 0

    raise NotImplementedError(f"neutral element for >= and {dtype}")


class MaxReductionOperation(ScalarReductionOperation):
    def neutral_element(self, dtype, callables_table, target):
        return get_ge_neutral(dtype), callables_table

    def __call__(self, dtype, operand1, operand2, callables_table, target):
        dtype, = dtype
        from loopy.translation_unit import add_callable_to_table

        # getting the callable 'max' from target
        max_scalar_callable = target.get_device_ast_builder().known_callables["max"]

        # type specialize the callable
        max_scalar_callable, callables_table = max_scalar_callable.with_types(
                {0: dtype, 1: dtype}, callables_table)

        # populate callables_table
        func_id, callables_table = add_callable_to_table(callables_table, "max",
                max_scalar_callable)

        return ResolvedFunction(func_id)(operand1, operand2), callables_table


class MinReductionOperation(ScalarReductionOperation):
    def neutral_element(self, dtype, callables_table, target):
        return get_le_neutral(dtype), callables_table

    def __call__(self, dtype, operand1, operand2, callables_table, target):
        dtype, = dtype
        from loopy.translation_unit import add_callable_to_table

        # getting the callable 'min' from target
        min_scalar_callable = target.get_device_ast_builder().known_callables["min"]

        # type specialize the callable
        min_scalar_callable, callables_table = min_scalar_callable.with_types(
                {0: dtype, 1: dtype}, callables_table)

        # populate callables_table
        func_id, callables_table = add_callable_to_table(callables_table, "min",
                min_scalar_callable)

        return ResolvedFunction(func_id)(operand1, operand2), callables_table


# {{{ base class for symbolic reduction ops

@expr_dataclass()
class ReductionOpFunction(FunctionIdentifier):
    reduction_op: ReductionOperation

    @property
    def name(self):
        return self.__class__.__name__

    def copy(self, reduction_op=None):
        if reduction_op is None:
            reduction_op = self.reduction_op

        return type(self)(reduction_op)

# }}}


# {{{ segmented reduction

class SegmentedOp(ReductionOpFunction):
    pass


class _SegmentedScalarReductionOperation(ReductionOperation):
    def __init__(self, **kwargs):
        self.inner_reduction = self.base_reduction_class(**kwargs)

    @property
    def base_reduction_class(self):
        raise NotImplementedError

    @property
    def which(self):
        raise NotImplementedError

    @property
    def arg_count(self):
        return 2

    def prefix(self, scalar_dtype, segment_flag_dtype):
        return "loopy_segmented_{}_{}_{}".format(self.which,
                scalar_dtype.numpy_dtype.type.__name__,
                segment_flag_dtype.numpy_dtype.type.__name__)

    def neutral_element(self, scalar_dtype, segment_flag_dtype,
            callables_table, target):
        from loopy.library.function import MakeTupleCallable
        from loopy.translation_unit import add_callable_to_table

        scalar_neutral_element, _calables_table = (
                self.inner_reduction.neutral_element(
                    scalar_dtype, callables_table, target))

        make_tuple_callable = MakeTupleCallable(
                name="make_tuple")

        make_tuple_callable, callables_table = make_tuple_callable.with_types(
                dict(enumerate([scalar_dtype, segment_flag_dtype])),
                callables_table)

        func_id, callables_table = add_callable_to_table(
                callables_table, "make_tuple", make_tuple_callable)

        return ResolvedFunction(func_id)(scalar_neutral_element,
                segment_flag_dtype.numpy_dtype.type(0)), callables_table

    def result_dtypes(self, scalar_dtype, segment_flag_dtype):
        return ((*self.inner_reduction.result_dtypes(scalar_dtype), segment_flag_dtype))

    def __str__(self):
        return "segmented(%s)" % self.which

    def __hash__(self):
        return hash(type(self))

    def __eq__(self, other):
        return type(self) is type(other) and (self.inner_reduction ==
                other.inner_reduction)

    def __call__(self, dtypes, operand1, operand2, callables_table, target):
        segmented_scalar_callable = SegmentOpCallable(SegmentedOp(self))

        # type specialize the callable
        segmented_scalar_callable, callables_table = (
                segmented_scalar_callable.with_types(
                    {0: dtypes[0], 1: dtypes[1], 2: dtypes[0], 3: dtypes[1]},
                    callables_table))

        # populate callables_table
        from loopy.translation_unit import add_callable_to_table
        func_id, callables_table = add_callable_to_table(
                callables_table, SegmentedOp(self), segmented_scalar_callable)

        return (ResolvedFunction(func_id)(*(operand1 + operand2)),
                callables_table)


class SegmentedSumReductionOperation(_SegmentedScalarReductionOperation):
    base_reduction_class = SumReductionOperation
    which = "sum"
    op = "((%s) + (%s))"

    hash_fields = (
            "which",
            "op",)

    update_persistent_hash = update_persistent_hash


class SegmentedProductReductionOperation(_SegmentedScalarReductionOperation):
    base_reduction_class = ProductReductionOperation
    op = "((%s) * (%s))"
    which = "product"

    hash_fields = (
            "which",
            "op",
            "base_reduction_class",)

    update_persistent_hash = update_persistent_hash

# }}}


# {{{ argmin/argmax

@expr_dataclass()
class ArgExtOp(ReductionOpFunction):
    pass


class _ArgExtremumReductionOperation(ReductionOperation):

    @property
    def which(self):
        raise NotImplementedError

    @property
    def neutral_sign(self):
        raise NotImplementedError

    def prefix(self, scalar_dtype, index_dtype):
        return "loopy_arg{}_{}_{}".format(self.which,
                scalar_dtype.numpy_dtype.type.__name__,
                index_dtype.numpy_dtype.type.__name__)

    def result_dtypes(self, scalar_dtype, index_dtype):
        return (scalar_dtype, index_dtype)

    def neutral_element(self, scalar_dtype, index_dtype, callables_table,
            target):
        scalar_neutral_func = (
                get_ge_neutral if self.neutral_sign < 0 else get_le_neutral)
        scalar_neutral_element = scalar_neutral_func(scalar_dtype)

        from loopy.library.function import MakeTupleCallable
        from loopy.translation_unit import add_callable_to_table
        make_tuple_callable = MakeTupleCallable(
                name="make_tuple")

        make_tuple_callable, callables_table = make_tuple_callable.with_types(
                dict(enumerate([scalar_dtype, index_dtype])),
                callables_table)

        # populate callables_table
        func_id, callables_table = add_callable_to_table(callables_table,
                                                         "make_tuple",
                                                         make_tuple_callable)

        return ResolvedFunction(func_id)(scalar_neutral_element,
                index_dtype.numpy_dtype.type(-1)), callables_table

    def __str__(self):
        return "arg" + self.which

    def __hash__(self):
        return hash(type(self))

    def __eq__(self, other):
        return type(self) is type(other)

    @property
    def arg_count(self):
        return 2

    def __call__(self, dtypes, operand1, operand2, callables_table, target):
        arg_ext_scalar_callable = ArgExtOpCallable(ArgExtOp(self))

        # type specialize the callable
        arg_ext_scalar_callable, callables_table = (
                arg_ext_scalar_callable.with_types(
                    {0: dtypes[0], 1: dtypes[1], 2: dtypes[0], 3: dtypes[1]},
                    callables_table))

        # populate callables_table
        from loopy.translation_unit import add_callable_to_table
        func_id, callables_table = add_callable_to_table(
                callables_table, ArgExtOp(self), arg_ext_scalar_callable)

        return (ResolvedFunction(func_id)(*(operand1 + operand2)),
                callables_table)


class ArgMaxReductionOperation(_ArgExtremumReductionOperation):
    which = "max"
    update_comparison = ">="
    neutral_sign = -1

    hash_fields = ("which",
            "update_comparison",
            "neutral_sign",)

    update_persistent_hash = update_persistent_hash


class ArgMinReductionOperation(_ArgExtremumReductionOperation):
    which = "min"
    update_comparison = "<="
    neutral_sign = +1

    hash_fields = ("which",
            "update_comparison",
            "neutral_sign",)

    update_persistent_hash = update_persistent_hash

# }}}


# {{{ reduction op registry

_REDUCTION_OPS = {
        "sum": SumReductionOperation,
        "product": ProductReductionOperation,
        "max": MaxReductionOperation,
        "min": MinReductionOperation,
        "any": AnyReductionOperation,
        "all": AllReductionOperation,
        "argmax": ArgMaxReductionOperation,
        "argmin": ArgMinReductionOperation,
        "segmented(sum)": SegmentedSumReductionOperation,
        "segmented(product)": SegmentedProductReductionOperation,
        }

_REDUCTION_OP_PARSERS = [
        ]


def register_reduction_parser(parser):
    """Register a new :class:`loopy.library.reduction.ReductionOperation`.

    :arg parser: A function that receives a string and returns
        a subclass of ReductionOperation.
    """
    _REDUCTION_OP_PARSERS.append(parser)


def parse_reduction_op(name: str) -> ReductionOperation | None:
    import re

    red_op_match = re.match(r"^([a-z]+)_([a-z0-9_]+)$", name)
    if red_op_match:
        op_name = red_op_match.group(1)

        if op_name in _REDUCTION_OPS:
            from warnings import warn
            warn("Reductions with forced result types are no longer supported. "
                    f"Encountered '{name}', which might be one.",
                    DeprecationWarning, stacklevel=1)
            return None

    if name in _REDUCTION_OPS:
        return _REDUCTION_OPS[name]()

    for parser in _REDUCTION_OP_PARSERS:
        result = parser(name)
        if result is not None:
            return result

    return None

# }}}


# {{{ reduction specific callables

class ReductionCallable(ScalarCallable):
    def with_types(self, arg_id_to_dtype, callables_table):
        scalar_dtype = arg_id_to_dtype[0]
        index_dtype = arg_id_to_dtype[1]
        result_dtypes = self.name.reduction_op.result_dtypes(scalar_dtype,  # pylint: disable=no-member
                index_dtype)

        new_arg_id_to_dtype = constantdict(arg_id_to_dtype).mutate()
        new_arg_id_to_dtype[-1] = result_dtypes[0]
        new_arg_id_to_dtype[-2] = result_dtypes[1]
        name_in_target = self.name.reduction_op.prefix(scalar_dtype,  # pylint: disable=no-member
                index_dtype) + "_op"

        return (self.copy(arg_id_to_dtype=new_arg_id_to_dtype.finish(),
                          name_in_target=name_in_target),
                callables_table)

    def with_descrs(self, arg_id_to_descr, callables_table):
        from loopy.kernel.function_interface import ValueArgDescriptor

        new_arg_id_to_descr = constantdict(arg_id_to_descr).mutate()
        new_arg_id_to_descr[-1] = ValueArgDescriptor()

        return (
                self.copy(arg_id_to_descr=new_arg_id_to_descr.finish()),
                callables_table)


class ArgExtOpCallable(ReductionCallable):

    def generate_preambles(self, target):
        op = self.name.reduction_op  # pylint: disable=no-member
        scalar_dtype = self.arg_id_to_dtype[-1]
        index_dtype = self.arg_id_to_dtype[-2]

        prefix = op.prefix(scalar_dtype, index_dtype)

        yield (prefix, """
        inline {scalar_t} {prefix}_op(
            {scalar_t} op1, {index_t} index1,
            {scalar_t} op2, {index_t} index2,
            {index_t} *index_out)
        {{
            if (op2 {comp} op1)
            {{
                *index_out = index2;
                return op2;
            }}
            else
            {{
                *index_out = index1;
                return op1;
            }}
        }}
        """.format(
                scalar_t=target.dtype_to_typename(scalar_dtype),
                prefix=prefix,
                index_t=target.dtype_to_typename(index_dtype),
                comp=op.update_comparison,
                ))

        return


class SegmentOpCallable(ReductionCallable):

    def generate_preambles(self, target):
        op = self.name.reduction_op  # pylint: disable=no-member
        scalar_dtype = self.arg_id_to_dtype[-1]
        segment_flag_dtype = self.arg_id_to_dtype[-2]
        prefix = op.prefix(scalar_dtype, segment_flag_dtype)

        yield (prefix, """
        inline {scalar_t} {prefix}_op(
            {scalar_t} op1, {segment_flag_t} segment_flag1,
            {scalar_t} op2, {segment_flag_t} segment_flag2,
            {segment_flag_t} *segment_flag_out)
        {{
            *segment_flag_out = segment_flag1 | segment_flag2;
            return segment_flag2 ? op2 : {combined};
        }}
        """.format(
                scalar_t=target.dtype_to_typename(scalar_dtype),
                prefix=prefix,
                segment_flag_t=target.dtype_to_typename(segment_flag_dtype),
                combined=op.op % ("op1", "op2"),
                ))

        return

# }}}

# vim: fdm=marker
